/* pconn: Connect to ports through other hosts */
/* Mark Kamichoff <prox@penguinhosting.net>    */

#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <signal.h>
#include <netdb.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <string.h>

#define SA struct sockaddr              /* Improves readability of code, eh */

int listeningfd, clientfd, datafd;

void signalpipe(void);          // Handles a broken pipe
void sig_chld(void);            // Handles SIGCHLD
void signalinterrupt(void);
void setupsignals(void);        // Sets up signal handlers
void syntax(void);
void setupsignals(void);
int bytepass(struct hostent *, char *, char *);

void syntax()
{
	printf("Syntax: pconn <local address> <local port> <remote address> <remote port>\n");
}

int main(int argc, char **argv)
{
	socklen_t len;
	struct sockaddr_in laddr, raddr;
	struct hostent *hptr, *listen_hptr;
	struct in_addr **listen_pptr;

	setupsignals();

        if(argc != 5)
        {
		syntax();
                exit(1);
        }

	if(!(((atoi(argv[2]) < 65537) && (atoi(argv[2]) > 0))) &&
		((atoi(argv[4]) < 65537) && (atoi(argv[4]) > 0)))
	{
		printf("Error: Invalid port number\n");
		syntax();
		exit(1);
	}

	/* Start setting up the listening sockaddr_in */
        bzero(&laddr, sizeof(laddr));	// Zeroes the structure
        laddr.sin_family = AF_INET;     // IP

	// If we specified a local listening address
	if(strcmp(argv[1],"0") != 0)
	{
		if((listen_hptr = gethostbyname(argv[1])) == NULL)
		{
			// This is fatal!
			fprintf(stderr, "Error: Couldn't resolve listen address of %s\n",
				argv[1]);
			exit(1);
		}
		else
		{
			listen_pptr = (struct in_addr **) listen_hptr->h_addr_list;
			memcpy(&laddr.sin_addr, *listen_pptr, sizeof(struct in_addr));
		}
	}
	else
	        laddr.sin_addr.s_addr = htonl(INADDR_ANY);

        laddr.sin_port = htons(atoi(argv[2]));	// Listening port
	/* Done with setting up the listening sockaddr_in */

	// Sets up the listening socket for incoming connections
        if(!(listeningfd = socket(AF_INET, SOCK_STREAM, 0)))
        {
                fprintf(stderr,"Error: Socket creation failed, giving up\n");
                exit(1);
        }

        // Calls bind to listen on the specific port
        if(bind(listeningfd, (SA *) &laddr, sizeof(laddr)) != 0)
        {
                fprintf(stderr,"Error: bind() returned error for port %s\n",
                        argv[2]);
                exit(1);
        }

        /* Set the socket to listen for a connection */

        if(listen(listeningfd, 5) != 0)
        {
                fprintf(stderr,"Error: listen() returned an error\n");
                exit(1);
        }
	printf("Debug: Service bound to port %d\n", ntohs(laddr.sin_port));

	while(1)
	{
		len = sizeof(raddr);
		if((clientfd = accept(listeningfd, (SA *) &raddr,
			&len)) < 0)
		{
			fprintf(stderr, "Error: accept() error, dumping nonexistant client\n");
			continue;
		}

                // Do a reverse-lookup on our client
                hptr = gethostbyaddr((char *) &raddr.sin_addr,
                        sizeof(raddr.sin_addr),
                        AF_INET);

		if(fork() == 0)		// Child
		{
			close(listeningfd);
			if(bytepass(hptr, argv[3], argv[4]) == -1)
				fprintf(stderr, "Warning: bytepass() returned abnormally\n");
			exit(0);
		}
		close(clientfd);
	}
	exit(0);
}
